#!/usr/bin/perl -w

# $Id: webwatch.pl,v 1.7 2020/04/05 04:19:33 pmdg Exp $

#########################################################################
# webwatch.pl
#
# basic web version of 'nsrwatch'.
#
# Key REST API elements we need to monitor:
# 
# === DEVICES ===
# /nwrestapi/v3/global/devices
#		count: X
#		devices: list of hashes
#			name
#			mediaFamily
#			mediaType
#			readOnly ("true" or "false")
#			status ("Enabled" or other)
#			volumeId
#			volumePool
#
#
# === VOLUMES ===
# /nwrestapi/v3/global/volumes
#		count: X
#		volumes: list of hashes
#			id
#			name
#			mounts
#			pool
#			written -> hash -> { unit, value }
#			recover -> hash -> { unit, value }
#
# === SESSIONS ===
# NB: Not just active sessions.
# /nwrestapi/v3/global/sessions
#		count: X
#		sessions: list of hashes
#			completed: true or false
#			clientHostname
#			protectionGroup
#			device
#			mode: Saving, or others
#			pool
#			saveSet
#			size -> hash -> { unit, value }
#			startTime
#			stopped: true or false
#			transferRate -> hash -> { unit, value }
#			volume
#
#########################################################################

#########################################################################
# Modules
#########################################################################
use strict;
use JSON;
use Data::Dumper;
use REST::Client;
use MIME::Base64;


#########################################################################
# Global Variables
#########################################################################

# Create a "monitor" user in NetWorker and setup the credentials in the
# username and password below.
my $user = "watcher";
my $pass = "Backup321_";

# Change the server to reference your NetWorker Server.
my $server = "orilla.turbamentis.int";
my $auth = encode_base64($user . ":" . $pass);

my $headers = { Accept => "application/json", 
		 	    Authorization => "Basic $auth"
		 	  };

# The Data Directory needs to be created, and writeable by the
# user account that will run webwatch.pl
my $dataDir = "/Users/pmdg/tmp/webwatch";
		 	  
# If you are using self-signed ceriticates, you'll need to keep
# the following line enabled.
$ENV{PERL_LWP_SSL_VERIFY_HOSTNAME} = 0;


#########################################################################
# Subroutines
#########################################################################

# in_list($expr,@list) returns true iff $expr is in list
sub in_list {
	return 0 if (@_+0 < 2);
	
	my $element = $_[0];
	return 0 if (!defined($element));	# Undefined expression? That's a paddlin'.
	shift @_;
	
	my @list = @_;
	return 0 if (!@_ || @_+0 == 0);	# Empty list? That's a paddlin'.
	
	$element = quotemeta($element);
	
	my @results = grep(/^$element$/,@list);
	return @results+0;
}

# in_gb($value,$unit) returns the given NetWorker unit
# combination in GB.
sub in_gb {
	return 0 if (@_+0 != 2);
	
	my $value = $_[0];
	my $unit = lc($_[1]);
	
	return $value if ($unit eq "gb");
	
	# else...
	if ($unit eq "b") {
		return $value / 1000 / 1000 / 1000;
	} elsif ($unit eq "kb") {
		return $value / 1000 / 1000;
	} elsif ($unit eq "mb") {
		return $value / 1000;
	} elsif ($unit eq "tb") {
		return $value * 1000;
	} elsif ($unit eq "pb") {
		return $value * 1000 * 1000;
	}
}

# in_mb($value,$unit) returns the given NetWorker transfer unit
# combination in MB.
sub in_mb {
	return 0 if (@_+0 != 2);
	
	my $value = $_[0];
	my $unit = lc($_[1]);
	
	return $value if ($unit eq "gb");
	
	# else...
	if ($unit eq "b/s") {
		return $value / 1000 / 1000;
	} elsif ($unit eq "kb/s") {
		return $value / 1000;
	} elsif ($unit eq "mb/s") {
		return $value;
	} elsif ($unit eq "gb/s") {
		return $value * 1000;
	} elsif ($unit eq "tb/s") {
		return $value * 1000;
	}
}

# write_data("type",@list) writes the passed list to the file nominated by the "type".
# it also handles writing a sentinel file to help avoid a race condition where we read
# data that's currently being written.
sub write_data {
	die "Dying! Oh the codemanity!\n" if (@_+0 < 1);
	
	my $type = $_[0];
	shift @_;
	my @data = @_;	# This can be an empty list, at which point we write an empty file.
	
	if ($type ne "volumes" && $type ne "devices" && $type ne "sessions") {
		die "Received type: $type\n\nShould be 'volumes', 'devices' or 'sessions'\n";
	}
	
	# Create the sentinel first.
	my $sentinelFile = $dataDir . "/writing.lck";
	if (open(SENTINEL,">$sentinelFile")) {
		print SENTINEL localtime(time) . "\n";
		close(SENTINEL);
	} else {
		die "Unable to write sentinel file: $sentinelFile\n";
	}
	
	# Now, write the data.
	my $dataFile = $dataDir . "/" . $type . ".data";
	if (open(DATA,">$dataFile")) {
		foreach my $line (@data) {
			my $tmp = $line;
			chomp $tmp;	# Just in case.
			print DATA "$tmp\n";
		}
		close(DATA);
	} else {
		# Try to unlink the sentinel file first.
		unlink($sentinelFile);
		die "Unable to create $dataFile\nCheck $dataDir, and delete $sentinelFile if necessary.\n";
	}
	
	# All done, delete the sentinel file.
	if (!unlink($sentinelFile)) {
		die "Unexpected results deleting sentinel file ($sentinelFile), unsafe to continue.\n";
	}
}

# get_volume_data() gets volume data and writes it to the volume file.
sub get_volume_data {
	my $client = REST::Client->new();
	$client->setHost("https://$server:9090");
	$client->GET(
		'/nwrestapi/v3/global/volumes',
		$headers);
	
	my $response = from_json($client->responseContent());
	my %volumes = %{$response};

	if (defined($volumes{count})) {
		print "\tProcessing " . $volumes{count} . " volumes\n";
	}
	
	my @volData = ();

	if (defined($volumes{"volumes"})) {
		#printf("%-15s%-25s%-10s%14s%14s\n","Volume ID","Label","Mounts","Written","Recovered");

		foreach my $vol (@{$volumes{"volumes"}}) {
			my $written = "";
			my $recovered = "";
		
			my %volume = %$vol;
		
			my $volID = (defined($volume{"id"})) ? $volume{"id"} : -1;
			my $volName = (defined($volume{"name"})) ? $volume{"name"} : "null";
			my $mounts = (defined($volume{"mounts"})) ? $volume{"mounts"} : -1;
			my $volPool = (defined($volume{"pool"})) ? $volume{"pool"} : "null";
		
			if (defined($volume{"written"})) {
				my %wr = %{$volume{"written"}};
				if (defined($wr{"value"}) && defined($wr{"unit"})) {
					my $val = $wr{"value"};
					my $uni = $wr{"unit"};
					$written = sprintf("%.2f GB",in_gb($val,$uni));
				}
			}
			if (defined($volume{"recover"})) {
				my %re = %{$volume{"recover"}};
				if (defined($re{"value"}) && defined($re{"unit"})) {
					my $val = $re{"value"};
					my $uni = $re{"unit"};
					$recovered = sprintf("%.2f GB",in_gb($val,$uni));
				}
			}
		
			if ($volID + $mounts > 0) {
				#printf("%-15s%-25s%-10d%14s%14s\n",$volID,$volName,$mounts,$written,$recovered);
				push(@volData,"$volID|||$volName|||$volPool|||$mounts|||$written|||$recovered\n");
			}
		}
	}
	
	if (@volData+0 > 0) {
		write_data("volumes",@volData);
	}
}

# get_session_data() retrieves and writes session data.
sub get_session_data {
	my $client = REST::Client->new();
	$client->setHost("https://$server:9090");
	$client->GET(
		'/nwrestapi/v3/global/sessions',
		$headers);
	
	my $response = from_json($client->responseContent());

	my %sessions = %{$response};
	if (defined($sessions{"count"})) {
		print "\tProcessing " . $sessions{"count"} . " recorded sessions\n";
	}

	my @sessionList = ();
	my $activeSessionCount = 0;
	if (defined($sessions{"sessions"})) {
		foreach my $sess (@{$sessions{"sessions"}}) {
			my %session = %$sess;
		
			my $completed = (defined($session{"completed"})) ? $session{"completed"} : -1;
			my $device = (defined($session{"device"})) ? $session{"device"} : "no device";
			my $mode = (defined($session{"mode"})) ? $session{"mode"} : "null";
			my $pool = (defined($session{"pool"})) ? $session{"pool"} : "null";
			my $saveset = (defined($session{"saveSet"})) ? $session{"saveSet"} : "no saveset";
			my $volume = (defined($session{"volume"})) ? $session{"volume"} : "null";
			my $startTime = (defined($session{"startTime"})) ? $session{"startTime"} : "null";
			my $stopped = (defined($session{"stopped"})) ? $session{"stopped"} : 0;
			my $client = (defined($session{"clientHostname"})) ? $session{"clientHostname"} : "";
			my $group = (defined($session{"protectionGroup"})) ? $session{"protectionGroup"} : "";
			my $sizeGB = 0;
			my $xferMB = 0;
		
			next if ($completed);
		
			# else...
			$activeSessionCount++;
			if ($completed == 0) {
				$completed = "Not Completed";
			} else {
				$completed = "Completed";
			}
		
			if (defined($session{"size"})) {
				my %sz = %{$session{"size"}};
				my $sV = $sz{"value"};
				my $sU = $sz{"unit"};
				$sizeGB = in_gb($sV,$sU);
			}
			if (defined($session{"transferRate"})) {
				my %xfer = %{$session{"transferRate"}};
				my $xV = $xfer{"value"};
				my $xU = $xfer{"unit"};
				$xferMB = in_mb($xV,$xU);
			}
		
			push(@sessionList,"$completed|||$mode|||$saveset|||$client|||$group|||$device|||$volume|||$startTime|||$stopped|||$xferMB|||$sizeGB")
		}
	}
	
	# We even write a blank list.
	write_data("sessions",@sessionList);
	
	if ($activeSessionCount) {
		my $plural = ($activeSessionCount == 1) ? "" : "s";
		print "\t\t$activeSessionCount active session$plural\n";
	} else {
		print "\t\tNo active sessions.\n";
	}
}

# get_device_data() retrieves and writes device data.
sub get_device_data {
	# Now let's take a peek at the devices
	my $client = REST::Client->new();
	$client->setHost("https://$server:9090");
	$client->GET(
		'/nwrestapi/v3/global/devices',
		$headers);
	
	my $response = from_json($client->responseContent());

	my %devices = %{$response};
	if (defined($devices{"count"})) {
		print "\tProcessing " . $devices{"count"} . " devices\n";
	}
	
	my @devData = ();

	if (defined($devices{"devices"})) {
		foreach my $device (@{$devices{"devices"}}) {
			my %dev = %$device;
		
			my $devName = (defined($dev{"name"})) ? $dev{"name"} : "null";
			my $mediaFamily = (defined($dev{"mediaFamily"})) ? $dev{"mediaFamily"} : "null";
			my $mediaType = (defined($dev{"mediaType"})) ? $dev{"mediaType"} : "null";
			my $readOnly = (defined($dev{"readOnly"})) ? $dev{"readOnly"} : "null";
			my $mountedVolume = (defined($dev{"mountedVolume"})) ? $dev{"mountedVolume"} : "null";
			my $lastMessage = (defined($dev{"message"})) ? $dev{"message"} : "null";
			my $status = (defined($dev{"status"})) ? $dev{"status"} : "null";
			my $volID = (defined($dev{"volumeId"})) ? $dev{"volumeId"} : "null";
			my $volPool = (defined($dev{"volumePool"})) ? $dev{"volumePool"} : "null";

			#printf("\n%-30s\n\t%-6s%-15s%-10s%-10s%-15s%-10s\n",$devName,$mediaFamily,$mediaType,$readOnly,$status,$volID,$volPool);
			#printf("\t%-21s%-60s\n",$mountedVolume,$lastMessage)
			push(@devData,"$devName|||$mediaFamily|||$mediaType|||$readOnly|||$status|||$volID|||$volPool|||$mountedVolume|||$lastMessage");
		}
	}
	
	if (@devData+0 > 0) {
		write_data("devices",@devData);
	}
}


#########################################################################
# Main
#########################################################################

my $execCount = 0;
while (1 == 1) {
	# Need to modify this so we only retrieve the volume data every 5 checks.
	print "Current time: " . localtime(time) . "\n";
	get_device_data();
	get_session_data();
	if ($execCount % 5 == 0) {
		get_volume_data();
	} else {
		print "\tSkip volume poll\n";
	}
	$execCount++;
	print "\t(Sleep 60)\n\n";
	sleep (60);
}

# Now let's take a peek at the sessions
